/*
 *
 *  Copyright (C) 2005 Mike Isely <isely@pobox.com>
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/device.h>   // for linux/firmware.h
#include <linux/firmware.h>
#include "pvrusb2-encoder.h"
#include "pvrusb2-i2c.h"
#include "pvrusb2-hdw-low.h"
#include "pvrusb2-hdw-internal.h"
#include "pvrusb2-debug.h"

static u32 pvr_tbl_emphasis [] = {
    [PVR2_CVAL_AUDIOEMPHASIS_NONE] = 0x0 << 12,
    [PVR2_CVAL_AUDIOEMPHASIS_50_15] = 0x1 << 12,
    [PVR2_CVAL_AUDIOEMPHASIS_CCITT] = 0x3 << 12,
};

static u32 pvr_tbl_srate[] = {
    [PVR2_CVAL_SRATE_48] =  0x01,
    [PVR2_CVAL_SRATE_44_1] = 0x00,
};

static u32 pvr_tbl_audiobitrate[] = {
    [PVR2_CVAL_AUDIOBITRATE_384] = 0xe << 4,
    [PVR2_CVAL_AUDIOBITRATE_320] = 0xd << 4,
    [PVR2_CVAL_AUDIOBITRATE_256] = 0xc << 4,
    [PVR2_CVAL_AUDIOBITRATE_224] = 0xb << 4,
    [PVR2_CVAL_AUDIOBITRATE_192] = 0xa << 4,
    [PVR2_CVAL_AUDIOBITRATE_160] = 0x9 << 4,
    [PVR2_CVAL_AUDIOBITRATE_128] = 0x8 << 4,
    [PVR2_CVAL_AUDIOBITRATE_112] = 0x7 << 4,
    [PVR2_CVAL_AUDIOBITRATE_96]  = 0x6 << 4,
    [PVR2_CVAL_AUDIOBITRATE_80]  = 0x5 << 4,
    [PVR2_CVAL_AUDIOBITRATE_64]  = 0x4 << 4,
    [PVR2_CVAL_AUDIOBITRATE_56]  = 0x3 << 4,
    [PVR2_CVAL_AUDIOBITRATE_48]  = 0x2 << 4,
    [PVR2_CVAL_AUDIOBITRATE_32]  = 0x1 << 4,
    [PVR2_CVAL_AUDIOBITRATE_VBR] = 0x0 << 4,
};

struct pvr2_cmd post_encoder_cmd [] = {
    PVR2_CMD(0,    "\x01\x03\x00\x00\x00\x00\x00\x44"),
    PVR2_CMD(0x40, "\x02\x00\x00\x00\x00\x00\x00\x44"),
    PVR2_CMD(0x40, "\x28\x00\x00\x00\x00\x00\x00\x44"),
    PVR2_CMD(0x40, "\x28\x00\x00\x00\x00\x00\x00\x54"),
    PVR2_CMD(0,    "\x01\x00\x00\x00\x00\x00\x00\x44"),
    PVR2_CMD(0, "")
};

static int pvr2_write_encoder_cmd (struct pvr2_hdw *hdw, u8 cmd,
                                   u32 *data, int dlen)
{
    int ret = 0, i, j;
    u8 d;

#ifdef notdef
    switch (dlen) {
    case 0:
	pvr_debug("ENCODER CMD %02x",cmd);
	break;
    case 1:
	pvr_debug("ENCODER CMD %02x: %02x",cmd,data[0]);
	break;
    case 2:
	pvr_debug("ENCODER CMD %02x: %02x %02x",cmd,data[0],data[1]);
	break;
    case 3:
	pvr_debug("ENCODER CMD %02x: %02x %02x %02x",
		  cmd,data[0],data[1],data[2]);
	break;
    case 4:
	pvr_debug("ENCODER CMD %02x: %02x %02x %02x %02x",
		  cmd,data[0],data[1],data[2],data[3]);
	break;
    case 5:
	pvr_debug("ENCODER CMD %02x: %02x %02x %02x %02x %02x",
		  cmd,data[0],data[1],data[2],data[3],data[4]);
	break;
    case 6:
	pvr_debug("ENCODER CMD %02x: %02x %02x %02x %02x %02x %02x",
		  cmd,data[0],data[1],data[2],data[3],data[4],data[5]);
	break;
    case 7:
	pvr_debug("ENCODER CMD %02x: %02x %02x %02x %02x %02x %02x %02x",
		  cmd,data[0],data[1],data[2],data[3],data[4],data[5],data[6]);
	break;
    case 8:
	pvr_debug("ENCODER CMD %02x: %02x %02x %02x %02x %02x %02x %02x %02x",
		  cmd,data[0],data[1],data[2],data[3],data[4],data[5],data[6],
		  data[7]);
	break;
    case 9:
	pvr_debug("ENCODER CMD %02x: %02x %02x %02x %02x %02x %02x %02x %02x"
		  " %02x",
		  cmd,data[0],data[1],data[2],data[3],data[4],data[5],data[6],
		  data[7],data[8]);
	break;
    case 10:
	pvr_debug("ENCODER CMD %02x: %02x %02x %02x %02x %02x %02x %02x %02x"
		  " %02x %02x",
		  cmd,data[0],data[1],data[2],data[3],data[4],data[5],data[6],
		  data[7],data[8],data[9]);
	break;
    default:
	pvr_debug("ENCODER CMD %02x: %02x %02x %02x %02x %02x %02x %02x %02x"
		  " %02x (%d more words)",
		  cmd,data[0],data[1],data[2],data[3],data[4],data[5],data[6],
		  data[7],data[8],dlen-9);
	break;
    }
#endif

    down(&hdw->ctl_sem);

    memset(hdw->cmd_buffer, 0, PVR2_CTL_BUFFSIZE);
    hdw->cmd_buffer[0] = 0x01;
    hdw->cmd_buffer[8] = cmd;
    hdw->cmd_buffer[24] = 0x06;
    for (i = 7, d = 0x44 ;  i <= 56 ; i+=7, d++)
        hdw->cmd_buffer[i] = d;
    for (i = 29, j = 0 ; j < min(4, dlen) ; i+= 7, j++)
        PVR2_DECOMPOSE_LE(hdw->cmd_buffer, i, data[j]);

    ret |= pvr2_send_request(hdw, hdw->cmd_buffer, 57, NULL, 0);

    memset(hdw->cmd_buffer, 0, PVR2_CTL_BUFFSIZE);
    hdw->cmd_buffer[0] = 0x01;
    for (i = 7, d = 0x4c;  i <= 56 ; i+=7, d++)
        hdw->cmd_buffer[i] = d;
    for (i = 1, j = 4 ; j < min(8, dlen) ; i+= 7, j++)
        PVR2_DECOMPOSE_LE(hdw->cmd_buffer, i, data[j]);

    ret |= pvr2_send_request(hdw, hdw->cmd_buffer, 57, NULL, 0);

    memset(hdw->cmd_buffer, 0, PVR2_CTL_BUFFSIZE);
    hdw->cmd_buffer[0] = 0x01;
    for (i = 7, d = 0x54;  i <= 28 ; i+=7, d++)
        hdw->cmd_buffer[i] = d;

    ret |= pvr2_send_request(hdw, hdw->cmd_buffer, 57, NULL, 0);

    ret |= pvr2_send_sequence(hdw, post_encoder_cmd);

    up(&hdw->ctl_sem);

    return ret;
}

static int pvr2_write_encoder_vcmd (struct pvr2_hdw *hdw, u8 cmd,
                                    int args, ...)
{
    va_list vl;
    int i;
    u32 data[12];

    va_start(vl, args);
    for (i = 0; i < args; i++) {
        data[i] = va_arg(vl, u32);
    }
    va_end(vl);

    return pvr2_write_encoder_cmd (hdw, cmd, data, args);
}

int pvr2_encoder_configure(struct pvr2_hdw *hdw)
{
    int ret = 0, audio, i;
    int is_ntsc = (hdw->controls[PVR2_CID_VIDEOSTANDARD].value ==
                   PVR2_CVAL_VIDEOSTANDARD_NTSC_M);
    int height = hdw->controls[PVR2_CID_VRES].value;
    int width = hdw->controls[PVR2_CID_HRES].value;
    int height_full = !hdw->controls[PVR2_CID_INTERLACE].value;

    pvr2_trace(PVR2_TRACE_ENCODER,"pvr2_encoder_configure");

    /* set stream output port. */
    ret |= pvr2_write_encoder_vcmd(hdw, 0xbb, 2, 0x01, 0x01);

    /* set the Program Index Information. We want I,P,B frames (max 400) */
    ret |= pvr2_write_encoder_vcmd(hdw, 0xc7, 2, 0x07, 0x0190);

    /* NOTE : windows driver sends some 0xdc */

/*
  ret |= pvr2_write_encoder_vcmd(hdw, 0xdc, 1, 5);
  ret |= pvr2_write_encoder_vcmd(hdw, 0xdc, 2, 3, 1);
  ret |= pvr2_write_encoder_vcmd(hdw, 0xdc, 1, 8);
*/

    /* Strange compared to ivtv data. */
//    ret |= pvr2_write_encoder_vcmd(hdw, 0xd6, 2, 0x0120, 0x0120);
//    ret |= pvr2_write_encoder_vcmd(hdw, 0xd6, 2, 0x0131, 0x0131);
    ret |= pvr2_write_encoder_vcmd(hdw, 0xd6, 2, 0xf0, 0xf0);

    /* setup firmware to notify us about some events (don't know why...) */
    ret |= pvr2_write_encoder_vcmd(hdw, 0xd5, 4, 0, 0, 0x10000000, 0xffffffff);

    /* set fps to 25 or 30 (1 or 0)*/
    ret |= pvr2_write_encoder_vcmd(hdw, 0x8f, 1, is_ntsc ? 0 : 1);

    /* set encoding resolution */
    ret |= pvr2_write_encoder_vcmd(hdw, 0x91, 2,
                                   (height_full ? height : (height / 2)),
                                   width);
    /* set encoding aspect ratio to 4:3 */
    ret |= pvr2_write_encoder_vcmd(hdw, 0x99, 1, 0x02);

    /* VBI */

    if (hdw->config == pvr2_config_vbi) {
        int lines = 2 * (is_ntsc ? 12 : 18);
        int size = (4*((lines*1443+3)/4)) / lines;
        ret |= pvr2_write_encoder_vcmd(hdw, 0xc8, 7, 0xbd05, 1, 4,
                                       0x25256262, 0x387f7f7f, lines , size);
//                                     0x25256262, 0x13135454, lines , size);
        /* select vbi lines */
#define line_used(l)  (is_ntsc ? (l >= 10 && l <= 21) : (l >= 6 && l <= 23))
        for (i = 2 ; i <= 24 ; i++){
            ret |= pvr2_write_encoder_vcmd(hdw, 0xb7, 5, i-1,
                                           line_used(i), 0, 0, 0);
            ret |= pvr2_write_encoder_vcmd(hdw, 0xb7, 5, (i-1) | (1 << 31),
                                           line_used(i), 0, 0, 0);
        }
    } else
        ret |= pvr2_write_encoder_vcmd(hdw, 0xb7, 5, 0xffffffff,0,0,0,0);

    /* set stream type, depending on resolution. */
    ret |= pvr2_write_encoder_vcmd(hdw, 0xb9, 1, height_full ?
                                   0x0a : 0x0b);
    /* set video bitrate */
    ret |= pvr2_write_encoder_vcmd(
        hdw, 0x95, 3,
        (hdw->controls[PVR2_CID_VBR].value ? 1 : 0),
        hdw->controls[PVR2_CID_AVERAGEVIDEOBITRATE].value,
        (u32) (hdw->controls[PVR2_CID_PEAKVIDEOBITRATE].value) / 400);
    /* setup GOP structure (GOP size = 0f or 0c, 3-1 = 2 B-frames) */
    ret |= pvr2_write_encoder_vcmd(hdw, 0x97, 2, is_ntsc ?  0x0f : 0x0c, 0x03);

    /* enable 3:2 pulldown */
    ret |= pvr2_write_encoder_vcmd(hdw, 0xb1, 1, 0);

    /* set GOP open/close property (open) */
    ret |= pvr2_write_encoder_vcmd(hdw, 0xc5, 1, 0);

    /* set audio stream properties 0x40b9? 0100 0000 1011 1001 */
    audio = (pvr_tbl_audiobitrate[hdw->controls[PVR2_CID_AUDIOBITRATE].value] |
             pvr_tbl_srate[hdw->controls[PVR2_CID_SRATE].value] |
             hdw->controls[PVR2_CID_AUDIOLAYER].value << 2 |
             (hdw->controls[PVR2_CID_AUDIOCRC].value ? 1 << 14 : 0) |
             pvr_tbl_emphasis[hdw->controls[PVR2_CID_AUDIOEMPHASIS].value]);

    ret |= pvr2_write_encoder_vcmd(hdw, 0xbd, 1, audio);

    /* set dynamic noise reduction filter to manual, Horiz/Vert */
    ret |= pvr2_write_encoder_vcmd(hdw, 0x9b, 2, 0, 0x03);

    /* dynamic noise reduction filter param */
    ret |= pvr2_write_encoder_vcmd(hdw, 0x9d, 2, 0, 0);

    /* dynamic noise reduction median filter */
    ret |= pvr2_write_encoder_vcmd(hdw, 0x9f, 4, 0, 0xff, 0, 0xff);

    /* spacial prefiler parameter */
    ret |= pvr2_write_encoder_vcmd(hdw, 0xa1, 2, 0x01, 0x01);

    /* initialize video input */
    ret |= pvr2_write_encoder_vcmd(hdw, 0xcd, 0);

    return ret;
}

int pvr2_encoder_start(struct pvr2_hdw *hdw)
{
    if (hdw->config == pvr2_config_vbi)
        return pvr2_write_encoder_vcmd(hdw, 0x81, 2, 0x01, 0x14);
    else if (hdw->config == pvr2_config_mpeg)
        return pvr2_write_encoder_vcmd(hdw, 0x81, 2, 0, 0x13);
    else
        return pvr2_write_encoder_vcmd(hdw, 0x81, 2, 0, 0x13);
}

int pvr2_encoder_stop(struct pvr2_hdw *hdw)
{
    if (hdw->config == pvr2_config_vbi)
        return pvr2_write_encoder_vcmd(hdw, 0x82, 3, 0x01, 0x01, 0x14);
    else if (hdw->config == pvr2_config_mpeg)
        return pvr2_write_encoder_vcmd(hdw, 0x82, 3, 0x01, 0, 0x13);
    else
        return pvr2_write_encoder_vcmd(hdw, 0x81, 3, 0x01, 0, 0x13);
}
