/*
 *  $Id: pvrusb2-hdw-low.c,v 1.9 2005/05/21 16:42:05 isely Exp $
 *
 *  Copyright (C) 2005 Mike Isely <isely@pobox.com>
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pvrusb2-hdw-low.h"
#include "pvrusb2-hdw-internal.h"
#include "pvrusb2-debug.h"

#define PVR2_CTL_WRITE_ENDPOINT  0x01
#define PVR2_CTL_READ_ENDPOINT   0x81

void pvr2_reset_ctl_endpoints(struct pvr2_hdw *hdw)
{
    if (!hdw->usb_dev) return;
    usb_settoggle(hdw->usb_dev, PVR2_CTL_WRITE_ENDPOINT & 0xf,
                  !(PVR2_CTL_WRITE_ENDPOINT & USB_DIR_IN), 0);
    usb_settoggle(hdw->usb_dev, PVR2_CTL_READ_ENDPOINT & 0xf,
                  !(PVR2_CTL_READ_ENDPOINT & USB_DIR_IN), 0);
    usb_clear_halt(hdw->usb_dev,
                   usb_rcvbulkpipe(hdw->usb_dev,
                                   PVR2_CTL_READ_ENDPOINT & 0x7f));
    usb_clear_halt(hdw->usb_dev,
                   usb_sndbulkpipe(hdw->usb_dev,
                                   PVR2_CTL_WRITE_ENDPOINT & 0x7f));
}

int pvr2_send_request(struct pvr2_hdw *hdw, void *data,
                      unsigned int dlen, void *res,
                      unsigned int rlen)
{
    int ret = 0, actual_dlen;
    unsigned int pipe;

    if (!hdw->flag_ok) {
	pvr2_trace(PVR2_TRACE_ERROR_LEGS,
                   "Attempted to execute control transfer when device not ok");
        return -EIO;
    }

    /* send data if any. */
    if(dlen && data){
        pipe = usb_sndbulkpipe(hdw->usb_dev, PVR2_CTL_WRITE_ENDPOINT);
        ret = usb_bulk_msg(hdw->usb_dev, pipe, data, dlen,
                           &actual_dlen, HZ / 4);
        if (ret){
            info("transaction failed, retrying");
            pvr2_reset_ctl_endpoints(hdw);
            /* retry */
            ret = usb_bulk_msg(hdw->usb_dev, pipe, data, dlen,
                               &actual_dlen, HZ / 4);
            if (ret) {
		pvr2_hdw_render_useless_unlocked(hdw);
                return 0;
	    }
        }
    }

    if(rlen == 0)
        return 0;

    pipe = usb_rcvbulkpipe(hdw->usb_dev, PVR2_CTL_READ_ENDPOINT);
    ret = usb_bulk_msg(hdw->usb_dev, pipe, res, rlen, &actual_dlen, HZ / 4);

    if (ret){
        info("transaction failed, retrying");
        pvr2_reset_ctl_endpoints(hdw);
        /* retry */
        ret = usb_bulk_msg(hdw->usb_dev, pipe, res, rlen,
                           &actual_dlen, HZ / 4);
        if (ret) {
	    pvr2_trace(PVR2_TRACE_ERROR_LEGS,
                       "transaction failed after 2 attempts");
	    pvr2_hdw_render_useless_unlocked(hdw);
	}
    }

    return ret;
}

/*
 * pvr2_send_sequence().
 *
 * executes a sequence of commands ended by a zero length one.
 *
 */

int pvr2_send_sequence(struct pvr2_hdw *hdw, struct pvr2_cmd *cmd)
{
    int ret = 0, i;

    i = 0;
    while ((cmd[i].length != 0) && (ret == 0)){
        memcpy(hdw->cmd_buffer, cmd[i].cmd, cmd[i].length);
        ret |= pvr2_send_request(hdw, hdw->cmd_buffer, cmd[i].length,
                                 hdw->cmd_buffer, cmd[i].ret);
        i++;
    }
    return ret;
}

int pvr2_write_register(struct pvr2_hdw *hdw, u16 reg, u32 data)
{
    int ret;

    down(&hdw->ctl_sem);

    hdw->cmd_buffer[0] = 0x04;  /* write register prefix */
    PVR2_DECOMPOSE_BE(hdw->cmd_buffer,1,data);
    hdw->cmd_buffer[5] = 0;
    hdw->cmd_buffer[6] = (reg >> 8) & 0xff;
    hdw->cmd_buffer[7] = reg & 0xff;


    ret = pvr2_send_request(hdw, hdw->cmd_buffer, 8, hdw->cmd_buffer, 0);

    up(&hdw->ctl_sem);

    return ret;
}

int pvr2_read_register(struct pvr2_hdw *hdw, u16 reg, u32 *data)
{
    int ret = 0;

    down(&hdw->ctl_sem);

    hdw->cmd_buffer[0] = 0x05;  /* read register prefix */
    hdw->cmd_buffer[1] = 0;
    hdw->cmd_buffer[2] = 0;
    hdw->cmd_buffer[3] = 0;
    hdw->cmd_buffer[4] = 0;
    hdw->cmd_buffer[5] = 0;
    hdw->cmd_buffer[6] = (reg >> 8) & 0xff;
    hdw->cmd_buffer[7] = reg & 0xff;

    ret |= pvr2_send_request(hdw, hdw->cmd_buffer, 8, hdw->cmd_buffer, 4);
    *data = *((u32 *)hdw->cmd_buffer) & 0xffff;

    up(&hdw->ctl_sem);

    return ret;
}

int pvr2_write_u16(struct pvr2_hdw *hdw, u16 data, int res)
{
    int ret;

    down(&hdw->ctl_sem);

    hdw->cmd_buffer[0] = (data >> 8) & 0xff;
    hdw->cmd_buffer[1] = data & 0xff;

    ret = pvr2_send_request(hdw, hdw->cmd_buffer, 2, hdw->cmd_buffer, res);

    up(&hdw->ctl_sem);

    return ret;
}

int pvr2_write_u8(struct pvr2_hdw *hdw, u8 data, int res)
{
    int ret;

    down(&hdw->ctl_sem);

    hdw->cmd_buffer[0] = data;

    ret = pvr2_send_request(hdw, hdw->cmd_buffer, 1, hdw->cmd_buffer, res);

    up(&hdw->ctl_sem);

    return ret;
}

/*
 * pvr_do_command()
 *
 * Executes a simple command.
 */

int pvr2_do_command(struct pvr2_hdw *hdw, struct pvr2_cmd *cmd)
{
    int ret;

    if (cmd == NULL)
        return -EIO;

    down(&hdw->ctl_sem);

    /* data command (cmd->cmd) is (often) statically allocated. For usb
     *  transfer we need dynamic memory.
     */
    memcpy(hdw->cmd_buffer, cmd->cmd, cmd->length);

    ret = pvr2_send_request(hdw, hdw->cmd_buffer, cmd->length,
                            hdw->cmd_buffer, cmd->ret);

    up(&hdw->ctl_sem);

    return ret;
}

void pvr2_hdw_render_useless_unlocked(struct pvr2_hdw *hdw)
{
    if (!hdw->flag_ok) return;
    pvr2_trace(PVR2_TRACE_INIT,"render_useless");
    hdw->flag_ok = 0;
    if (hdw->vid_stream) {
        pvr2_stream_setup(hdw->vid_stream,0,0);
    }
    hdw->flag_streaming_enabled = 0;
}


void pvr2_hdw_render_useless(struct pvr2_hdw *hdw)
{
    down(&hdw->ctl_sem);
    pvr2_hdw_render_useless_unlocked(hdw);
    up(&hdw->ctl_sem);
}


void pvr2_hdw_cpureset_assert(struct pvr2_hdw *hdw,int val)
{
    char da[1];
    unsigned int pipe;
    int ret;

    if (!hdw->usb_dev) return;

    pvr2_trace(PVR2_TRACE_INIT,"cpureset_assert(%d)",val);

    da[0] = val ? 0x01 : 0x00;

    /* Write the CPUCS register on the 8051.  The lsb of the register is
       the reset bit; a 1 asserts reset while a 0 clears it. */
    pipe = usb_sndctrlpipe(hdw->usb_dev, 0);
    ret = usb_control_msg(hdw->usb_dev,pipe,0xa0,0x40,0xe600,0,da,1,HZ);
    if (ret < 0) {
        pvr2_trace(PVR2_TRACE_ERROR_LEGS,
                   "cpureset_assert(%d) error=%d",val,ret);
        pvr2_hdw_render_useless(hdw);
    }
}


/*
  Stuff for Emacs to see, in order to encourage consistent editing style:
  *** Local Variables: ***
  *** mode: c ***
  *** fill-column: 75 ***
  *** c-basic-offset: 4 ***
  *** indent-tabs-mode: nil ***
  *** End: ***

  Similarly for vim to see:
  vim:expandtab:textwidth=75

*/
